/*
	Program for generating drawings of the triangulations, that have
	nearly the same area of all the faces

	Written by Anicka Bernathova <anicka@anicka.net> in 2009
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <math.h>

#include <cairo.h>
#include <cairo-pdf.h>
#include <fcntl.h>
#include <unistd.h>

void __attribute__ ((noreturn))
die(char* msg, ...)
{
        va_list args;

        va_start(args, msg);
        vfprintf(stderr, msg, args);
        fputc('\n', stderr);
        va_end(args);
        exit(1);
}

void*
xmalloc(size_t size)
{
        void* ret;

        if (!(ret = malloc(size)))
                die("Low memory");

        return ret;
}

#define BUFSIZE 1024
#define MAX_VERTICES 100
#define MAX_FACES (MAX_VERTICES/3) 
#define MAX_SHEEP 1000 /* size of the population, bigger is slower but yields better results */
#define MAX_SHEEP_DRAW 100 /* set this to one if you want one big drawing of a graph */

struct coord {
	int v;
	double x;
	double y;
}; /* structure for drawing of one vertex */

struct vertex {
	int v;	
	int n_neigh;
	int neighbours[MAX_VERTICES];
	int v_faces[MAX_FACES];
};

struct sheep {
	struct coord c[MAX_VERTICES];
	double rank;
	int id;
	int p1_id;
	int p2_id;
}; /* here we hold all the information about one member of our population */

int f[MAX_FACES][3]; /* list of faces */
struct vertex graph[MAX_VERTICES]; /* the graph we are working with */
int n_vertices; /* number of vertices of our graph */
int nfaces; /* number of its faces */
struct sheep shed[MAX_SHEEP]; /* our population */
int gen; /* current generation number */
int current_id; /* id of the next sheep born */
struct coord start_coord[MAX_VERTICES]; /* the drawing we start with */
int read_faces; /* used during reading input */
#ifdef DEBUG
FILE* logfile;
#endif

/*
 As we want to produce well proportioned pictures, we normalize all our drawing so
 as their maximum x and y coordinates maps to 1.
*/
void
normalize(struct coord* c)
{
	double x_max = -HUGE_VAL, x_min = HUGE_VAL, y_max = -HUGE_VAL,
		 y_min = HUGE_VAL;
	int i;

	for (i = 0; i < n_vertices; i++) {
		if (c[i].x > x_max)
			x_max = c[i].x;	
		if (c[i].y > y_max)
			y_max = c[i].y;	
		if (c[i].x < x_min)
			x_min = c[i].x;	
		if (c[i].y < y_min)
			y_min = c[i].y;	
	}
	x_max -= x_min;
	y_max -= y_min;
	
	for (i = 0; i < n_vertices; i++) {
		c[i].x -= x_min;
		c[i].y -= y_min;

		c[i].x *= 1 / x_max;
		c[i].y *= 1 / y_max;
	}
	
}

/* read one line of the input from stdin */
void
read_line(char* s)
{
	int rc = 0, c, r, i = 0;

	if (*s == '\n') {
		read_faces++;
		return;
	}

	if (read_faces) {
		r = sscanf(s, "%d %d %d", &f[nfaces][0], &f[nfaces][1],
			&f[nfaces][2]);
		if (r != 3)
			die("parse error: r = %d, nfaces = %d", r, nfaces);
		nfaces++;
		return;
	}

	r = sscanf(s, "%lf %lf %n", &start_coord[n_vertices].x, &start_coord[n_vertices].y, &c);
	rc += c;
	if (r != 2)
		die("parse error, scanf strasi");
	while ((r = sscanf(s + rc,"%d %n",
		&graph[n_vertices].neighbours[i],&c)) == 1) {
		rc += c;
		i++;
	}
	graph[n_vertices].n_neigh = i;
	graph[n_vertices].v = n_vertices;
	start_coord[n_vertices].v = n_vertices;
	n_vertices++;
}

/* calculate are of one triangular face using the determinant method */
double
get_area(double x1, double y1, double x2, double y2, double x3, double y3)
{
	x2 -= x1;
	x3 -= x1;
	y2 -= y1;
	y3 -= y1;

	return ((x2 * y3) - (x3 * y2)) / 2.0;
}

/* calculate rank for one drawing */
double
get_solution(struct coord* c)
{
	double avg = 0, a, res = 0;
	int i;

	for (i = 0; i < nfaces; i++) {
		avg += fabs(get_area(c[f[i][0]].x, c[f[i][0]].y,
				c[f[i][1]].x, c[f[i][1]].y,
				c[f[i][2]].x, c[f[i][2]].y)); 

	}
	avg /= nfaces;

	for (i = 0; i < nfaces; i++) {
		a = fabs(get_area(c[f[i][0]].x, c[f[i][0]].y,
                             c[f[i][1]].x, c[f[i][1]].y,
                             c[f[i][2]].x, c[f[i][2]].y));
		res += (a - avg) * (a - avg);
	}

	return res / (avg * avg);
}

/* return oriented distance of the point from the line */
double
get_halfplane(struct coord p, struct coord l1, struct coord l2)
{
	double a, b, c, norm;

	a = -l1.y + l2.y;
	b = l1.x - l2.x;
	norm = sqrt(a*a + b*b);
	a /= norm;
	b /= norm;
	c = - ((a * l1.x) + (b * l1.y));

	return (a*p.x) + (b*p.y) + c;
}

/* check that the drawing of the vertex did not cross any drawing of an edge */
int
check_vertex(struct coord* co0, struct coord* cv, double nx, double ny)
{
	double r1, r2, ar_new;
	int i;
	struct coord c1, c2;
	struct coord nc = { .x = nx, .y = ny}; 

	for (i = 0; i < nfaces; i++) {
		int* face = f[i];

		if (!graph[cv->v].v_faces[i])
			continue;
		if (face[0] == cv->v) {
			c1 = co0[face[1]]; 
			c2 = co0[face[2]]; 
		} else if (face[1] == cv->v) {
			c1 = co0[face[0]]; 
			c2 = co0[face[2]]; 
		} else {
			c1 = co0[face[1]]; 
			c2 = co0[face[0]]; 
		}

		r1 = get_halfplane(*cv, c1, c2);
		r2 = get_halfplane(nc, c1, c2);
		ar_new = get_area(nx, ny, c1.x, c1.y, c2.x, c2.y);

#ifdef MORE_CHECKS
		double ar_orig = 
			get_area(cv->x, cv->y, c1.x, c1.y, c2.x, c2.y);
		if (ar_orig * ar_new < 0)
			return 1;
#endif
		if (fabs(ar_new) < 0.01)
			return 1;
		if (fabs(r1) < 0.01 || fabs(r2) < 0.01 || (r1*r2) < 0)
			return 1;
	}
	return 0;
}

/* get a random offset for moving a drawing of the vertex */
double
move(void)
{
	int r = random() % 1000;
	double res = (1 / (double) (r + 10));

	if (random() % 2)
		return res;
	else
		return -res;

}

void
get_v_faces(struct vertex* gr)
{
	int i;

	memset(gr->v_faces, 0, sizeof(int) * MAX_FACES);
	for (i = 0; i < nfaces; i++) {
		if ((f[i][0] == gr->v) || (f[i][1] == gr->v)
			|| (f[i][2] == gr->v))
			gr->v_faces[i]++;
	}
}

void
first_fill_shed(void)
{
	int i;
	double rank = get_solution(start_coord);	
	for (i = 0; i < MAX_SHEEP; i++){
		memcpy(shed[i].c, start_coord, 
			MAX_VERTICES * sizeof(struct coord));
		shed[i].rank = rank;
		shed[i].id = 0;
	}	
}

int
cmp(const void* p1, const void* p2)
{
	struct sheep* s1 = (struct sheep*) p1;
	struct sheep* s2 = (struct sheep*) p2;

	if (s1->rank == s2->rank)
		return 0;
	if (s1->rank < s2->rank)
		return -1;
	else 
		return 1;	
}

void
sort_shed(void)
{
	qsort(shed, MAX_SHEEP, sizeof(struct sheep), cmp);
}

/* pick one drawing from the population, better with greater probability */
int
pick_sheep(int sheep_max)
{
	int sum = 0;
	int b1 = sheep_max / 10, b2 = (sheep_max / 10) * 4;
	int r, i, j, h;
	int v[sheep_max];

	for (i = 0; i < sheep_max; i++) {
		if (i < b1)
			sum += 100;	
		else if (i < b2)
			sum += 10;
		else
			sum += 1; 
		v[i] = sum;
	}

	r = rand() % sum;
	i = 0; j = sheep_max - 1;

	while (j - i > 1) {
		h = i + ((j - i) / 2); 
		if (r < v[h])
			j = h;
		else 
			i = h;	
	}
	
	return i;	
}

/* create a new drawing from some two old ones (or clone one if the breeding does not go well) */
struct sheep
breed_sheep(void)
{
	int s1, s2;
	int i, j;
	struct sheep new;

	s1 = pick_sheep(MAX_SHEEP - (MAX_SHEEP / 20)); 
	s2 = pick_sheep(MAX_SHEEP - (MAX_SHEEP / 20)); 

	for (i = 0; i < n_vertices; i++) {
		new.c[i].x = (shed[s1].c[i].x + shed[s2].c[i].x) / 2;
 		new.c[i].y = (shed[s1].c[i].y + shed[s2].c[i].y) / 2;
		new.c[i].v = i;
	}

	for (i = 0; i < nfaces; i++) {
		for (j = 0; j < 3; j++) {
			struct coord nc, nl1, nl2, oc, ol1, ol2;
			double nr, or;
			int* face = f[i];

			nc = new.c[face[j]];
			nl1 = new.c[face[(j + 1)%3]];		
			nl2 = new.c[face[(j + 2)%3]];		

			oc = shed[s1].c[face[j]];
			ol1 = shed[s1].c[face[(j + 1)%3]];		
			ol2 = shed[s1].c[face[(j + 2)%3]];		
		
			nr = get_halfplane(nc, nl1, nl2);
			or = get_halfplane(oc, ol1, ol2);

			if (fabs(nr) < 0.01 || 
				fabs(or) < 0.01 || nr * or < 0) {
				return shed[s1];
			}
			
		}
	}
	normalize(new.c);
	new.rank = get_solution(new.c);
	new.id = current_id++;
	new.p1_id = shed[s1].id;
	new.p2_id = shed[s2].id;

	return new;
}

/* take care for killing bad drawings and breeding the population first, doing
 * random mutations in it and ranking it then */
void
oh_my_god(void)
{
	int i;
	int check, rv;
	double newx, newy;

	for (i = MAX_SHEEP - (MAX_SHEEP / 20); i < MAX_SHEEP; i++) {
		shed[i] = breed_sheep(); 
	}
	for (i = 0; i < MAX_SHEEP; i++) {
		do {	
			do {
				rv = random() % n_vertices;
				newx = shed[i].c[rv].x + move();	
				newy = shed[i].c[rv].y + move();
				check = check_vertex(shed[i].c, 
					shed[i].c + rv, newx, newy);
			} while (check);
			shed[i].c[rv].x = newx; shed[i].c[rv].y = newy;
		} while (random() % 10); 
		normalize(shed[i].c);
		shed[i].rank = get_solution(shed[i].c);
	}
}

/* draw 100 or less best drawings to PDF (or only one big picture if
 * MAX_SHEEP_DRAW is set to 1 */
void
draw_gen_top100(struct sheep* s, char* name)
{
	int i, j, sh, max;
	int diffx, diffy;
	struct coord* c;
	int m = 10, t = 100;
	cairo_surface_t *surface;
        cairo_t *cr;

        surface = cairo_pdf_surface_create(name, 1120, 1120);
        cr = cairo_create (surface);
        cairo_set_source_rgb (cr, 0xff, 0xff, 0xff);
        cairo_paint (cr);

	if (MAX_SHEEP_DRAW == 1) {
		m = 100;
		t = 1000;
	}
		
	max = MAX_SHEEP_DRAW < 100 ? MAX_SHEEP_DRAW : 100;
	for (sh = 0; sh < max; sh++) {
		c = s[sh].c;
		for (i = 0; i < n_vertices; i++) {
			for (j = 0; j < graph[i].n_neigh; j++) {
				diffx = (sh % m) * t + (sh % m) * m;
				diffy = (sh / m) * t + (sh / m) * m; 
				if (graph[i].neighbours[j] > i)
					cairo_set_source_rgb (cr, 0, 0, 0);
					cairo_move_to (cr,  t * c[i].x + diffx + 10, t * c[i].y + diffy + 10);
					cairo_line_to (cr, t * c[graph[i].neighbours[j]].x + diffx + 10, 
							t * c[graph[i].neighbours[j]].y + diffy + 10);
					cairo_set_line_width (cr, 1);
					cairo_stroke (cr);
			}
		}
	}
	cairo_destroy(cr);
        cairo_surface_destroy(surface);
}

/* save information about best drawings */
void
save_gen_top100(struct sheep* s, char* name)
{
	int i, last;
	double a;
	struct coord* c = s->c;
	FILE* out = fopen(name, "w");

	if (!out)
		die("%m");

	fputs("Best sheep:\n", out);
	for (i = 0; i < nfaces; i++) {
		a = get_area(c[f[i][0]].x, c[f[i][0]].y,
				c[f[i][1]].x, c[f[i][1]].y,
				c[f[i][2]].x, c[f[i][2]].y); 
		fprintf(out, "\tArea of face %d: %lf \n", i, a);
	}
		fprintf(out, "\t rank: %lf\n", s->rank);

	last = MAX_SHEEP < 100 ? MAX_SHEEP - 1 : 99;
	c = s[last].c;	
	fputs("100th sheep:\n", out);
	for (i = 0; i < nfaces; i++) {
		a = get_area(c[f[i][0]].x, c[f[i][0]].y,
				c[f[i][1]].x, c[f[i][1]].y,
				c[f[i][2]].x, c[f[i][2]].y); 
		fprintf(out, "\tArea of face %d: %lf \n", i, a);
	}
		fprintf(out, "\t rank: %lf\n", s[last].rank);
	
	fclose(out);
}

int
set_seed(void)
{
	int fd;
	int res;

	fd = open("/dev/urandom", O_RDONLY);
	if (fd < 0)
		return 1;
	read(fd, &res, sizeof(int));

	return res;
}

int
main(int argc, char** argv)
{
	char s[BUFSIZE];	
	int genmax = 1000, skip = 1;
	int i; 
	double best_rank = 1000;
	char namepdf[BUFSIZE];
	char nametxt[BUFSIZE];
	int seed;

#ifdef DEBUG
	logfile = fopen("log", "w");
	if (!logfile)
		die("%m");
#endif

	if (argc > 1)
		genmax = atoi(argv[1]);
	if (argc > 2)
		skip = atoi(argv[2]);
	if (argc > 3)
		seed = atoi(argv[3]);
	else
		seed = set_seed();
#ifdef DEBUG
	fprintf(stderr, "seed: %d\n", seed);	
#endif
	srand(seed);
	while (fgets(s, BUFSIZE, stdin)) {
		read_line(s);
	}

	for (i = 0; i < MAX_VERTICES; i++)
		get_v_faces(graph + i);
	normalize(start_coord);
	first_fill_shed();

	while(!genmax || (gen < genmax)) {
		sort_shed();

#ifdef DEBUG
		fprintf(logfile, "gen %d:\n", gen);
		for (i = 0; i < MAX_SHEEP; i++) {
			fprintf(logfile, "%d: id: %d (%d, %d)\n", i, shed[i].id,
			shed[i].p1_id, shed[i].p2_id);
		}
		fprintf(logfile, "\n");
#endif		
		if (!skip) {
			if (shed->rank < best_rank) {
				best_rank = shed->rank;
				printf("generation %d, best rank: %lf\n", gen, best_rank);
				goto draw;
			} else
				goto god;
		}

		if (gen % skip)
			goto god;
draw:
		sprintf(namepdf, "gen%.10d.pdf", gen);
		sprintf(nametxt, "gen%.10d.txt", gen);
		draw_gen_top100(shed, namepdf);
		save_gen_top100(shed, nametxt);
god:
		oh_my_god();
		gen++;
	}

#ifdef DEBUG
	printf("vertices: %d\n", n_vertices);
	fclose(logfile);
#endif

	return 0;
}
